###############################################################
# File "SaveToFiles.R":
# Set of functions creating the files that are input to
# tool 3D-VisualCluster.
#
#  Jos A. Castellanos G.
#  University of Valladolid
#  Department of Computer Sciences
#  Last update: 20/10/2011
###############################################################

print("File SaveToFiles.R:")
print("Set of functions creating the files that are input to tool 3D-VisualCluster.")
print("Jos A. Castellanos G.")
print("University of Valladolid")
print("Department of Computer Science")
print("Last update: 20/10/2011")

###############################################################
# CreateFiles function, creates three files from the output dendrogram
# returned by one of the hierarchical clustering methods built in R.
# That is, a dendrogram of classes "twins", "hclust", "agnes", etc.
# Then, these files have a format that can be loaded by the
# 3D-VisualCluster tool.
#
# INPUT:
#
# datamat: a data matrix representing the used dataset.
# agn: an dendrogram in R of class twins, the output of a hierarchical
#      clustering method.
# namedataset: name given to the dataset without the extension. This
#              name will be used to identify the three created files.
# part: it is an optional parameter that if used, then it means,
#       the part of the dendrogram that we want to remove (a valor in
#       form of fraction). This is usefull when we do not want to show
#       all dendrogram levels. For example, if we do not want to show
#       the first half of a dendrogram, then "part" have to be assigned
#       to 1/2 (part = 1/2). Hence, the first half of the dendrogram
#       will be removed.
#
# EXAMPLE IN R OF "CreateFiles":
#
# > data(votes.repub)
# > dendrogram <- diana(votes.repub, metric = "manhattan", stand = TRUE)
# > CreateFiles(votes.repub, dendrogram, "VotesRepub_Diana_Manhattan", part = 3/4)
#
# This creates the files "VotesRepub_Diana_Manhattan50x31.txt",
# "VotesRepub_Diana_Manhattan50x31_dendo.txt" and
# "VotesRepub_Diana_Manhattan50x31_dendo_gr.txt" in the current directory.
#
# Note that "part = 3/4" removes the first 3/4 part of the dendrograma,
# that is, since "dendrogram" has 50 levels, the first 39 levels are
# removed (50 * 3/4 + 2 = 39) and just the 11 remaining levels (50 - 39)
# are used.  On the other hand, if we want to explore the whole dendrogram,
# it is enough do not use "part", namely:
# CreateFiles(votes.repub, dendrogram, "VotesRepub_Diana_Manhattan")
###############################################################

CreateFiles <- function(datamat, agn, dataset.name, part = 2){
  numrow <- nrow(datamat); numcol <- ncol(datamat)
  dataset.name <- paste(dataset.name, numrow, "x", numcol, sep="")
  # It creates a file with "datamat" as a dataset, includes column
  # and row heads.
  write.table(datamat, file= paste(dataset.name,".txt", sep=""), sep =";",
              row.names = paste(c("g"), 1:numrow, sep=""),
              col.names = paste(c("c"), 1:numcol, sep=""), quote = F)
  # It transforms "agn" to the 3D-VisualCluster structure.
  dendo <- transf.tree(agn, part)
  dendo <- pos.mergedcluster(dendo)
  # It prints to a file all levels of the "dendo" dendrogram.
  write(paste(dataset.name,".txt", sep=""),
        file = paste(dataset.name, "_dendo", ".txt", sep=""))
  sink(paste(dataset.name, "_dendo", ".txt", sep=""), append = TRUE)
  print(dendo)
  sink()
  # It prints to a file the heights and merges of the "agn" dendrogram.
  sink(paste(dataset.name, "_dendo_gr", ".txt", sep=""))
  print(agn$height)
  print(agn$merge)
  sink()
  # Print information for the user.
  print("The following files have been created in the current directory: ")
  print(paste(dataset.name,".txt", sep=""))
  print(paste(dataset.name, "_dendo", ".txt", sep=""))
  print(paste(dataset.name, "_dendo_gr", ".txt", sep=""))
}

#############################################################
# The Create.RefPartitionFile function converts a clustering (which 
# is a reference partition) given by a vector structure (internal 
# structure of R)  into a clustering in form of list (3D-VisualCluster 
# structure) and saves the list to a file.
#
# INPUT:
#
# vector.clustering: reference partition (or clustering) as a vector of R.
# filename: name of the file (without extension) where the reference 
#           partition will be saved.
#
# OUTPUT: it saves and returns a clustering in form of list of R. 
#
# EXAMPLE:
#
# > vector.clustering <- c(1,2,1,3,2,3,4,4,5,1,2,5)
# > Create.RefPartitionFile(vector.clustering, "RefPart_Test")
#
# [1] "File RefPart_Test.txt was created in the current directory."
# [[1]]
# [1]  1  3 10
# [[2]]
# [1]  2  5 11
# [[3]]
# [1] 4 6
# [[4]]
# [1] 7 8
# [[5]]
# [1]  9 12
# [[6]]
# [1] 0 0
#############################################################

Create.RefPartitionFile <- function(vector.clustering, filename){
  extract.cluster <- function(i)
    list.clustering[[i]] <<- which(vector.clustering == i)
    
  cluster.indexes <- unique(vector.clustering)
  list.clustering <- list()
  
  sapply(cluster.indexes, extract.cluster, USE.NAMES = F)
  list.clustering <- c(list.clustering, list(c(0,0)))
  
  # It prints to a file the reference partition in list.clustering.
  sink(paste(filename, ".txt", sep=""))
  print(list.clustering)
  sink()
  
  # Print information for the user.
  print(paste("File ", filename, ".txt was created in ",
              "the current directory.", sep=""))
              
  return(list.clustering)
}


#############################################################
#
# Auxiliary functions just for use of function CreateFiles.
#
#############################################################

transf.tree <- function(agn, part = 2){
  setlevels <- function(i){
    tree[[n - i]] <<- transftoclusters(cutree(agn, k = size), orderv, size)
    size <<- size - 1
  }  
 
  n <- length(agn$order)
  orderv <- agn$order
  tree <- list(NULL)
  if (part != 2)  part <- trunc(n * part) + 2
  size <- n - part + 1
  sapply((n - 1):part, setlevels, USE.NAMES = F)

  return (tree)
}

transftoclusters <- function(vector.clust, orderdata, k){
  ordercluster <- function(clustering, orderdata, k){
     setclust <- function(i){
        elem <- orderdata[pos]; j <- 1
        while (j <= k)
          if (elem %in% clustering[[j]]) break()
            else j <- j + 1
        len <- length(clustering[[j]])
        tempclust[[i]] <<- orderdata[pos:(pos + len - 1)]
        pos <<- pos + len
     }
 
     tempclust <- list(NA); pos <- 1
     sapply(1:k, setclust, USE.NAMES = F)
     
     return (tempclust)
  }

  createcluster <- function(i)
    clustering[[i]] <<- which(vector.clust == i)
     
  clustering <- list(NA)
  vector.clust <- as.vector(vector.clust)
  sapply(1:k, createcluster, USE.NAMES = F)
  clustering <- ordercluster(clustering, orderdata, k)
  clustering[[k + 1]] <- c(0, 0)
  
  return (clustering)
}

# Function indentifying the positions of the clusters 
# that are merged to built the next level in the dendrogram.
pos.mergedcluster <- function(dendo)
{
 update.clustering <- function(i){
   k <- 0; j <- 1; sizeclust <<- sizeclust - 1
   while ((j <= sizeclust) && (k < 2)){
     if (!cluster.in.clust(dendo[[i]][[j]], 
                           dendo[[i + 1]], sizeclust - 1))
       {
        k <- k + 1
        dendo[[i + 1]][[sizeclust]][k] <<- j 
       }
     j <- j + 1 
  }  
 }
 
 sizedendo <- length(dendo)
 sizeclust <- length(dendo[[1]]) 
  
 sapply(1:(sizedendo - 1), update.clustering, USE.NAMES = F)
 
 return (dendo)   
}
 
cluster.in.clust <- function(cluster1, clustering, long)
{
 subcluster <- function(cluster1, cluster2)
  all(cluster1 %in% cluster2) && all(cluster2 %in% cluster1)
 
 i <- 1                                
 # Searching a cluster in a clustering.
 while (i <= long){
  if (subcluster(cluster1, clustering[[i]])) break()
  i <- i + 1
 } 
 return (i <= long)         
}
